local AddonName, GGL = ...
--[[-----------------------------------------------------------------------------
Label Widget
Displays text and optionally an icon.
-------------------------------------------------------------------------------]]
local Type, Version = "GoldGoblinLootGUIWidget-CountItemLine", 1
local AceGUI = LibStub and LibStub("AceGUI-3.0", true)
if not AceGUI or (AceGUI:GetWidgetVersion(Type) or 0) >= Version then return end

-- Lua APIs
local max, select, pairs = math.max, select, pairs
local tonumber = tonumber
local abs, floor = math.abs, math.floor
local format = string.format

-- WoW APIs
local CreateFrame, UIParent = CreateFrame, UIParent
local GetItemIcon, GetItemInfo = GetItemIcon, GetItemInfo

--[[-----------------------------------------------------------------------------
Support functions
-------------------------------------------------------------------------------]]
local GAP_SIZE = 2
local FIX_HEIGHT = 20

local GetMoneyText = GGL.GetMoneyTextShort

local function Frame_OnEnter(self)
    if not self.obj.itemID then return end
    local tooltip = _G.GameTooltip
    tooltip:ClearLines()
    GGL.GUI:SetToolTipPoint(tooltip, self.obj.ttParent, self.obj.ttPoint, 2, 0)
    tooltip:SetItemByID(self.obj.itemID)
    tooltip:Show()
end

local function Frame_OnLeave(self)
    _G.GameTooltip:Hide()
end

--[[-----------------------------------------------------------------------------
Methods
-------------------------------------------------------------------------------]]
local methods = {
    ["OnAcquire"] = function(self)
        self.itemID = nil
        self.value = 0
        self.amount = 0
        self.moneyValue = 0
        self.sortByMoney = true
        self:ResetToolTipPoint()
        self:EnableBG(false)
	end,

    ["OnWidthSet"] = function(self, width)

    end,

    ["SetName"] = function(self, name)
        self.name:SetText(name)
    end,

    ["SetToolTipPoint"] = function(self, ttPoint, parent)
        self.ttPoint = ttPoint or "BOTTOM_X"
        self.ttParent = parent or self.frame
    end,

    ["ResetToolTipPoint"] = function(self)
        self.ttPoint = "BOTTOM_X"
        self.ttParent = self.frame
    end,

    ["UpdateBG"] = function(self)
        if self.enableBG and self.itemID then
            self.bg:Show()
            local _, _, quality = GetItemInfo(self.itemID)
            local r,g,b = GetItemQualityColor(quality)
            self.bg:SetColorTexture(r,g,b,0.3)
        else
            self.bg:Hide()
        end
    end,

    ["EnableBG"] = function(self, state)
        self.enableBG = state
        self:UpdateBG()
    end,

    -- @return      addedMoney
    ["SetAmount"] = function(self, amount, replace, instantUpdate)
        if replace then
            self.amount = amount or 0
        else
            self.amount = (self.amount or 0) + amount
        end
        self.moneyValue = self.amount * self.value
        if instantUpdate then
            self.count:SetText(self.amount)
            self.money:SetText(GetMoneyText(self.moneyValue))
        end
        self:UpdateSortValue()
        return amount * self.value
    end,

    ["SetMoneyValue"] = function(self, value, instantUpdate)
        self.value = value or 0
        if instantUpdate then
            self:SetAmount(self.amount, true, instantUpdate)
        else
            self:UpdateSortValue()
        end
    end,

    ["SetItem"] = function(self, itemID)
        if not itemID then return end
        self.itemID = itemID

        self:SetMoneyValue(0, true)
        self:SetAmount(0, true, true)

        self.imageL:SetTexture(GetItemIcon(itemID))
        self:UpdateBG()
    end,

    ["SortByMoney"] = function(self, state)
        self.sortByMoney = state
        self:UpdateSortValue()
    end,

    ["UpdateSortValue"] = function(self)
        if self.sortByMoney then
            self:SetSortValue(self.moneyValue)
        else
            self:SetSortValue(self.amount)
        end
    end,

    ["SetSortValue"] = function(self, value)
        self.sortValue = value or 0
    end,

    ["GetSortValue"] = function(self)
        return self.sortValue
    end,

    ["UpdateListRank"] = function(self, newRank, listUpdate)
        self:SetAmount(self.amount, true, true)
    end,
}

--[[-----------------------------------------------------------------------------
Constructor
-------------------------------------------------------------------------------]]
local function Constructor()
    local widget = {}
    widget.type  = Type

    local frame = CreateFrame("Frame", nil, UIParent)
    frame:SetHeight(FIX_HEIGHT)
    frame:SetScript("OnEnter", Frame_OnEnter)
    frame:SetScript("OnLeave", Frame_OnLeave)
    frame.obj = widget
    widget.frame = frame

    local bg = frame:CreateTexture(nil, "BACKGROUND", nil, 1)
    bg:SetAllPoints(frame)
    widget.bg = bg

    local imageL = frame:CreateTexture(nil, "BACKGROUND", nil, 2)
    imageL:SetPoint("LEFT", frame, "LEFT", 0, 0)
    imageL:SetWidth(FIX_HEIGHT-2)
    imageL:SetHeight(FIX_HEIGHT-2)
    widget.imageL = imageL

    local count = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall", 2)
    count:SetPoint("LEFT", imageL, "RIGHT", GAP_SIZE, 0)
    count:SetJustifyH("LEFT")
    count:SetJustifyV("MIDDLE")
    count:SetWordWrap(false)
    count:SetHeight(FIX_HEIGHT)
    count:SetText(99999)
    count:SetWidth(count:GetStringWidth()+1)
    count:SetText(0)
    widget.count = count

    local line = frame:CreateTexture(nil, "BACKGROUND", nil, 2)
    line:SetPoint("TOPLEFT", count, "TOPRIGHT", 0, 0)
    line:SetPoint("BOTTOMLEFT", count, "BOTTOMRIGHT", 0, 0)
    line:SetWidth(1)
    line:SetColorTexture(1, 1, 1, 0.5)

    --[[
    local name = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall")
    name:SetPoint("LEFT", imageL, "RIGHT", GAP_SIZE, 0)
    name:SetPoint("RIGHT", frame, -1, 0)
    name:SetJustifyH("LEFT")
    name:SetJustifyV("MIDDLE")
    name:SetWordWrap(true)
    name:SetHeight(FIX_HEIGHT)
    widget.name = name
    ]]--

    local money = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightSmall", 2)
    money:SetPoint("LEFT", line, "RIGHT", 0, 0)
    money:SetPoint("RIGHT", frame, -1, 0)
    money:SetJustifyH("RIGHT")
    money:SetJustifyV("MIDDLE")
    money:SetWordWrap(false)
    widget.money = money

    -- methodes
    for method, func in pairs(methods) do
        widget[method] = func
    end

    return AceGUI:RegisterAsWidget(widget)
end

AceGUI:RegisterWidgetType(Type, Constructor, Version)
