local AddonName, GGL = ...

local LootAlert = {}
GGL.LootAlert = LootAlert
local L = GGL.Locale

local LSM = LibStub:GetLibrary("LibSharedMedia-3.0")

-- lua
local tbl_insert, tbl_remove = table.insert, table.remove

-- wow
local PlaySoundFile = PlaySoundFile

-- other
local OPERATORS_LIST, OPERATORS_DEFAULT = GGL.OPERATORS_LIST, GGL.OPERATORS_DEFAULT

local GetMoneyText = GGL.GetMoneyText
--local GetMoneyTextShort = GGL.GetMoneyTextShort

-- ########################################
-- Base
-- ########################################
local RAID_WARN_COLOR = {r=1.0, g=1.0, b=1.0}
local RAID_WARN_FORMAT, RAID_WARN_FORMAT_PLAYER = "%s %s", "%s %s (%s)"

local NOTIFY_TYPES = {
    {   name = L["None"] },
    {
        name = L["GGL Loot Alert"],
        method = function(itemID, itemString, value, playerTemplate)
            LootAlert:AddNewAlertItem(itemID, itemString, value, playerTemplate)
        end,
    },
    {
        name = L["Raid warning"],
        method = function(itemID, itemString, value, playerTemplate)
            if not itemID then return end

            local itemName, itemLink = GetItemInfo(itemString or itemID)
            if playerTemplate then
                _G.RaidNotice_AddMessage( _G.RaidWarningFrame, format(RAID_WARN_FORMAT_PLAYER, itemLink, GetMoneyText(value or 0), playerTemplate.displayName), RAID_WARN_COLOR, 2.5 )
            else
                _G.RaidNotice_AddMessage( _G.RaidWarningFrame, format(RAID_WARN_FORMAT, itemLink, GetMoneyText(value or 0)), RAID_WARN_COLOR, 2.5 )
            end
        end,
    },
    --[[ HW protected out of instances
    {   name = L["Chat message - Say"],
        method = function(itemID, value, playerTemplate)
            SendChatMessage(itemID, "SAY")
        end,
    },
    ]]
}
LootAlert.NOTIFY_TYPES = NOTIFY_TYPES

local function CheckOperatorValue(value, operator, valueCheck)
    operator = OPERATORS_LIST[operator or OPERATORS_DEFAULT]
    if not value or not valueCheck then return end

    -- switch ;_;
    if operator == "<" then
        return value < valueCheck
    elseif operator == ">" then
        return value > valueCheck
    elseif operator == "<=" then
        return value <= valueCheck
    elseif operator == ">=" then
        return value >= valueCheck
    elseif operator == "==" then
        return value == valueCheck
    elseif operator == "~=" then
        return value ~= valueCheck
    end
end

-- ########################################
-- func
-- ########################################
local dbTable, isGlobalRule

local function Init()
    dbTable = GGL.db.AlertRules

    -- Register sounds
    LSM:Register("sound", "GGL: AuctionWindowOpen", 567482)
    LSM:Register("sound", "GGL: AuctionWindowClose", 567499)
    -- Dasinka requests
    LSM:Register("sound", "GGL: IgQuestListComplete", 567439)
    LSM:Register("sound", "GGL: PVPFlagTakenHordeMono", 567427)
    LSM:Register("sound", "GGL: PVPWarning", 567488)
    LSM:Register("sound", "GGL: AlarmClockWarning2", 567399)
    LSM:Register("sound", "GGL: LFG_Rewards", 567514)
    LSM:Register("sound", "GGL: UI_PetBattle_PVP_Victory", 648423)

    LootAlert:UpdateAlertAnchor()
end
GGL:AddInitFuncPlayerLogin(Init)

--- Process alerts
-- @param       itemID              <number>
-- @para        itemString          <item:string>
-- @param       value               <number> copper value of an item
-- @param       playerTemplate      <playerTemplate>
function LootAlert:ProcessAlerts(itemID, itemString, value, playerTemplate)
    if self.paused or dbTable.disabled or not itemID then return end

    local itemName, itemLink, itemRarity, _ = GetItemInfo(itemID)

    for i = 1, #dbTable.Rules do
        local rule = dbTable.Rules[i]

        if not rule.disabled and (not isGlobalRule or GGL.PriceSource:IsValidID(rule.priceSource)) then
            if isGlobalRule then
                if rule.useVariations then
                    _, value = GGL.PriceSource:GetFunction(rule.priceSource)(itemID, itemString)
                else
                    value = GGL.PriceSource:GetFunction(rule.priceSource)(itemID)
                end
            end
            if (rule.isGlobal or isGlobalRule or (playerTemplate and playerTemplate.isPlayer)) and
            (CheckOperatorValue(itemRarity, rule.rarityOperator, rule.rarity or 0) and
            CheckOperatorValue(value, rule.goldOperator, rule.gold or 0)) then

                -- play sound
                if rule.sound and rule.sound ~= "None" and LSM:IsValid("sound", rule.sound) then
                    PlaySoundFile(LSM:Fetch("sound", rule.sound), "master")
                end

                -- run notify
                if NOTIFY_TYPES[rule.notify] and NOTIFY_TYPES[rule.notify].method then
                    NOTIFY_TYPES[rule.notify].method(itemID, itemString, value, playerTemplate)
                end

                -- end here if break is set
                if not rule.noBreak then
                    return
                end
            end
        end
    end
end

--- Pause the loot alert process
-- @param       state       <bool>
function LootAlert:Pause(state)
    if state then
        self.paused = true
    else
        self.paused = nil
    end
end

-- ########################################
-- Global Alerts
-- ########################################

local function GlobalAlert_OnItemLoot(Loot, itemID, itemLog, ammount, source, sourceID, mainSourceFlag, subSourceFlag, categoryFlag)
    if GGL.db.GlobalAlertRules.disabled or (GGL.db.GlobalAlertRules.disabledInChallenge and GGL.Session:GetLockedSession()) then return end
    dbTable = GGL.db.GlobalAlertRules
    isGlobalRule = true
    local newItemID, itemString = GGL:GetItemDataFromItemLog(itemLog)

    LootAlert:ProcessAlerts(newItemID, itemString, 0, nil)

    dbTable = GGL.db.AlertRules
    isGlobalRule = false
end

local function InitGlobalAlerts()
    GGL.Loot:RegisterCallbackEvent("Loot_OnItemLoot", GlobalAlert_OnItemLoot)
end
GGL:AddInitFuncPlayerLogin(InitGlobalAlerts)

-- ########################################
-- AlertFrame
-- ########################################

-- AlertFrameManager
do
    local FRAME_WIDTH, FRAME_HEIGHT, FRAME_GAP = 250, 55, 2

    local AlertFrameAnchor = CreateFrame("FRAME", nil, UIParent)
    AlertFrameAnchor:SetSize(FRAME_WIDTH, 20)
    AlertFrameAnchor:SetClampedToScreen(true)

    -- ## Anchor
    local function AlertFrameAnchor_OnMouseDown(self)
        AlertFrameAnchor:StartMoving()
    end

    local function AlertFrameAnchor_OnMouseUp(self)
        AlertFrameAnchor:StopMovingOrSizing()
        local point, relativeTo, relativePoint, xOfs, yOfs = AlertFrameAnchor:GetPoint()

        GGL.db.AlertNotify.AlertFrame.frame = {point, nil, relativePoint, xOfs, yOfs}
    end

    --- Update the anchor position
    function LootAlert:UpdateAlertAnchor()
        local points = GGL.db.AlertNotify.AlertFrame.frame
        AlertFrameAnchor:ClearAllPoints()
        AlertFrameAnchor:SetPoint(points[1], points[2] or UIParent, points[3], points[4], points[5])
    end

    --- Lock/Unlock the anchor frame
    -- @param       state       <bool>
    function LootAlert:UnlockAlertFrameAnchor(state)
        -- nil just flip the state
        if state == nil then
            state = not self.isUnlocked
        end
        if state then
            if self.isUnlocked then return end

            if not AlertFrameAnchor.bg then
                local bg = AlertFrameAnchor:CreateTexture(nil, "BACKGROUND")
                bg:SetAllPoints(AlertFrameAnchor)
                bg:SetColorTexture(0, 1, 0, 0.7)

                AlertFrameAnchor.bg = bg
            end

            AlertFrameAnchor.bg:Show()

            AlertFrameAnchor:SetScript("OnMouseDown", AlertFrameAnchor_OnMouseDown)
            AlertFrameAnchor:SetScript("OnMouseUp", AlertFrameAnchor_OnMouseUp)
            AlertFrameAnchor:SetMovable(true)

            self.isUnlocked = true
        else
            if not self.isUnlocked then return end

            if AlertFrameAnchor.bg then
                AlertFrameAnchor.bg:Hide()
            end

            AlertFrameAnchor:SetScript("OnMouseDown", nil)
            AlertFrameAnchor:SetScript("OnMouseUp", nil)
            AlertFrameAnchor:SetMovable(false)

            self.isUnlocked = false
        end
    end

    -- ## framePool
    local framePool = {}
    local activeFrames = {}

    local function GetFrameFromPool()
        local frame = next(framePool)
        if frame then
            framePool[frame] = nil
            return frame
        end
    end

    local function ReanchorActiveFrames()
        for i = 1, #activeFrames do
            if i == 1 then
                activeFrames[i]:SetPoint("BOTTOM", AlertFrameAnchor, "TOP", 0, 0)
            else
                activeFrames[i]:SetPoint("BOTTOM", activeFrames[i-1], "TOP", 0, 2)
            end
        end
    end

    local function FreeFrame(frame)
        frame:Hide()
        for i = 1, #activeFrames do
            if activeFrames[i] == frame then
                tbl_remove(activeFrames, i)
                break
            end
        end
        framePool[frame] = true

        ReanchorActiveFrames()
    end


    local function AlertFrame_PauseOutAnimation(self)
        self.waitAndAnimOut:Stop()
    end

    local function AlertFrame_PlayOutAnimation(self, optionalDelay)
        self.waitAndAnimOut.animOut:SetStartDelay(optionalDelay or 1)
        self.waitAndAnimOut:Play()
    end

    local function AlertFrame_ResumeOutAnimation(self)
        AlertFrame_PlayOutAnimation(self, 1)
    end

    local function AlertFrame_PlayIntroAnimation(self)
        self.animIn:Play()
    end

    local function AlertFrame_PlayOutroAnimation(self)
        AlertFrame_PauseOutAnimation(self)

        if not self:IsMouseOver() then
            AlertFrame_PlayOutAnimation(self, self.duration or 4)
        end
    end

    local function AlertFrame_PlayAnimations(self)
        AlertFrame_PlayIntroAnimation(self)

        AlertFrame_PlayOutroAnimation(self)
    end

    local function AlertFrame_OnEnter(self)
        self:PauseOutAnimation()

        if not self.itemID and not self.itemString then return end

        local tooltip = _G.GameTooltip
        tooltip:ClearLines()
        tooltip:SetOwner(UIParent,"ANCHOR_CURSOR")
        if self.itemString then
            tooltip:SetHyperlink(self.itemString)
        else
            tooltip:SetItemByID(self.itemID)
        end
        tooltip:Show()
    end

    local function AlertFrame_OnLeave(self)
        self:ResumeOutAnimation()
        _G.GameTooltip:Hide()
    end

    local function AlertFrame_OnMouseDown(self)
        LootAlert:UnlockAlertFrameAnchor(true)
        AlertFrameAnchor_OnMouseDown(AlertFrameAnchor)
    end

    local function AlertFrame_OnMouseUp(self)
        LootAlert:UnlockAlertFrameAnchor(false)
        AlertFrameAnchor_OnMouseUp(AlertFrameAnchor)
    end

    local function AnimOut_OnFinished(self)
        FreeFrame(self:GetRegionParent())
    end

    local function GetFrame()
        local frame = GetFrameFromPool()
        if not frame then
            frame = GGL.GUI:CreateBorderedFrame(AlertFrameAnchor)
            frame:SetSize(FRAME_WIDTH, FRAME_HEIGHT)
            frame:SetScript("OnEnter", AlertFrame_OnEnter)
            frame:SetScript("OnLeave", AlertFrame_OnLeave)
            frame:SetScript("OnMouseDown", AlertFrame_OnMouseDown)
            frame:SetScript("OnMouseUp", AlertFrame_OnMouseUp)

            -- methods
            frame.PauseOutAnimation = AlertFrame_PauseOutAnimation
            frame.PlayOutAnimation = AlertFrame_PlayOutAnimation
            frame.ResumeOutAnimation = AlertFrame_ResumeOutAnimation
            frame.PlayIntroAnimation = AlertFrame_PlayIntroAnimation
            frame.PlayOutroAnimation = AlertFrame_PlayOutroAnimation
            frame.PlayAnimations = AlertFrame_PlayAnimations

            -- #########################
            -- fadeAdnimation START
            local animIn = frame:CreateAnimationGroup()
            frame.animIn = animIn
            animIn.aList = {}
            animIn.aList[1] = animIn:CreateAnimation("Alpha")
            animIn.aList[1]:SetFromAlpha(1)
            animIn.aList[1]:SetToAlpha(0)
            animIn.aList[1]:SetDuration(0)
            animIn.aList[1]:SetOrder(1)

            animIn.aList[2] = animIn:CreateAnimation("Alpha")
            animIn.aList[2]:SetFromAlpha(0)
            animIn.aList[2]:SetToAlpha(1)
            animIn.aList[2]:SetDuration(0.2)
            animIn.aList[2]:SetOrder(2)

            local waitAndAnimOut = frame:CreateAnimationGroup()
            frame.waitAndAnimOut = waitAndAnimOut
            waitAndAnimOut.animOut = waitAndAnimOut:CreateAnimation("Alpha")
            waitAndAnimOut.animOut:SetStartDelay(4.05)
            waitAndAnimOut.animOut:SetFromAlpha(1)
            waitAndAnimOut.animOut:SetToAlpha(0)
            waitAndAnimOut.animOut:SetDuration(1.5)
            waitAndAnimOut.animOut:SetScript("OnFinished", AnimOut_OnFinished)
            -- fadeAdnimation END
            -- #########################

            local content = CreateFrame("Frame", nil, frame)
            content:SetPoint("TOPLEFT", frame, 5, -5)
            content:SetPoint("BOTTOMRIGHT", frame, -5, 5)
            content.obj = frame
            frame.content = content

            -- itemIcon
            local itemIcon = content:CreateTexture(nil, "BACKGROUND")
            itemIcon:SetPoint("TOPLEFT", content, "TOPLEFT", 0, 0)
            itemIcon:SetSize(FRAME_HEIGHT-10, FRAME_HEIGHT-10)
            frame.itemIcon = itemIcon

            -- itemNameText
            local itemNameText = content:CreateFontString(nil, "BACKGROUND", "GameFontHighlight")
            itemNameText:SetPoint("TOP", content, "TOP", 0, 0)
            itemNameText:SetPoint("LEFT", itemIcon, "RIGHT", FRAME_GAP, 0)
            itemNameText:SetPoint("RIGHT", content, "RIGHT", -FRAME_GAP, 0)
            itemNameText:SetHeight(15)
            itemNameText:SetJustifyH("LEFT")
            itemNameText:SetJustifyV("MIDDLE")
            frame.itemNameText = itemNameText

            -- goldText
            local goldText = content:CreateFontString(nil, "BACKGROUND", "GameFontHighlight")
            goldText:SetPoint("TOPLEFT", itemNameText, "BOTTOMLEFT", 0, 0)
            goldText:SetPoint("RIGHT", itemNameText, "RIGHT", 0, 0)
            goldText:SetHeight(15)
            goldText:SetJustifyH("LEFT")
            goldText:SetJustifyV("MIDDLE")
            frame.goldText = goldText

            -- playerText
            local playerText = content:CreateFontString(nil, "BACKGROUND", "GameFontHighlight")
            playerText:SetPoint("TOPLEFT", goldText, "BOTTOMLEFT", 0, 0)
            playerText:SetPoint("RIGHT", goldText, "RIGHT", 0, 0)
            playerText:SetHeight(15)
            playerText:SetJustifyH("LEFT")
            playerText:SetJustifyV("MIDDLE")
            frame.playerText = playerText


        end
        return frame
    end

    --- Adds a new alert item
    -- @param       itemID              <number>
    -- @param       itemString          <item:string>
    -- @param       value               <number> copper value of an item
    -- @param       playerTemplate      <playerTemplate>
    function LootAlert:AddNewAlertItem(itemID, itemString, value, playerTemplate)
        if not itemID then return end

        -- get a new alertFrame
        local alertFrame = GetFrame()
        alertFrame:Show()

        -- set itemID
        alertFrame.itemID = itemID
        alertFrame.itemString = itemString

        -- get item data
        local itemName, itemLink, itemRarity = GetItemInfo(itemString or itemID)

        -- set item data
        alertFrame.itemNameText:SetText(itemLink or UNKNOWN)
        alertFrame.itemIcon:SetTexture(GetItemIcon(itemID))
        alertFrame.goldText:SetText(GetMoneyText(value or 0))
        alertFrame.playerText:SetText(playerTemplate and playerTemplate.displayName or "")

        -- start animation
        alertFrame:PlayAnimations()

        -- check for active frames and reanchor them if needed
        if #activeFrames < 1 then
            activeFrames[1] = alertFrame
            alertFrame:SetPoint("BOTTOM", AlertFrameAnchor, "TOP", 0, 0)
        else
            tbl_insert(activeFrames, 1, alertFrame)
            -- reanchor
            ReanchorActiveFrames()
        end
    end
end
