-- ###########################
-- EVENTS:
--	"Session_OnSessionLocked" - Session, sessionTemplate
--  "Session_OnSessionLockReleased" - Session, sessionTemplate
-- ###########################

local AddonName, GGL = ...

local Session = {}
GGL.Session = Session

local AceGUI = LibStub("AceGUI-3.0")
local L = GGL.Locale

-- lua
local pairs = pairs
local match, strsub, format = string.match, strsub, string.format

-- wow

-- other
local Templates = GGL.Templates
local GetPlayerFromSync = Templates.Sync.GetPlayerFromSync

-- callback handler
GGL:AddCallbackHandler(Session)

-- ###########################
-- local
-- ###########################
local SessionTypes = {
    -- Challenge
    ["C"] = {
        key = "C",
        name = "Challenge",
        static = Templates.Challenge,
    },

    -- Multibox
    ["M"] = {
        key = "M",
        name = "Multibox",
        static = Templates.Multibox,
    },

    -- Session
    ["S"] = {
        key = "S",
        name = "Session",
        static = Templates.Session,
    },

    -- Solo
    ["A"] = {
        key = "A",
        name = "Solo",
        static = Templates.Solo,
    },
}

local SPECIAL_EVENT_DD_LIST = {
    Opening = L["Opening"]
}

-- return sessionID, sessionType
local function FormatSessionData(sessionData)
    local sessionID, sessionType
    if SessionTypes[sessionData] then
        sessionType = sessionData
    else
        -- check if we have a ID
        sessionType = strsub(sessionData, 1, 1)
        if SessionTypes[sessionType] then
            sessionID = sessionData
        else
            -- no valid sessionData
            return nil, sessionType
        end
    end
    return sessionID, sessionType
end

local function GetSessionFromSessionData(sessionData)
    local session
    if type(sessionData) == "string" then
        session = Session:TryGetHostSession(sessionData)
    elseif type(sessionData) == "table" then
        session = sessionData
    else
        return false
    end
    return session
end

-- ###########################
-- base
-- ###########################
local StartedSession
local SessionRegister = {}

local function InitLoop()
    if not Templates.BNWoWPlayer:GetSelfPlayer() then
        C_Timer.After(0.5, InitLoop)
        return
    end

    -- load running session but dont resume just stop it
    if GGL.sessionDb.RunningSession then
        local startedSession = Session:GetNewSession(GGL.sessionDb.RunningSession)
        startedSession:Stop()
        GGL.sessionDb.RunningSession = nil
    end

    -- recreate all sessions from list but stop them
    for sessionType, sessionID in pairs(GGL.sessionDb.SessionRegister) do
        local session = Session:GetNewSession(sessionID)
        -- they should not be running but just in case :D
        session:Stop()
    end
end

local function Init()
    InitLoop()
end
GGL:AddInitFuncPlayerLogin(Init)

--- Gets a session with the player as host
-- @param       sessionData         <string> sessionType or sessionID
-- @return      session (if false then there is a running session and the player is not the host)
function Session:TryGetHostSession(sessionData)
    if not sessionData then return end

    -- format sessionData
    local sessionID, sessionType = FormatSessionData(sessionData)

    -- started session found
    if StartedSession and (not StartedSession.PlayerIsValidHost or StartedSession:PlayerIsValidHost()) and StartedSession:GetSessionType() == sessionType then
        return StartedSession
    elseif StartedSession then
        return false
    end

    -- check if there is one in the register
    -- there should be only playerHost entrys in the SessionRegister table but better check
    if SessionRegister[sessionType] and not SessionRegister[sessionType].isClosed and (not SessionRegister[sessionType].PlayerIsValidHost or SessionRegister[sessionType]:PlayerIsValidHost()) then
        return SessionRegister[sessionType]
    else
        return self:GetNewSession(sessionData)
    end
end

--- Get a running session for a sessionType if possible
-- @param       sessionData         <string> sessionType or sessionID
-- @param       closeRunning        <bool> close the running session, if false the running only gets stopped
-- @return  sessionTemplate
function Session:GetNewSession(sessionData, closeRunning)
    if not sessionData then return end

    -- format sessionData
    local sessionID, sessionType = FormatSessionData(sessionData)

    if not sessionType then return end

    if closeRunning then
        self:CloseSession(sessionType)
    else
        self:StopSession(sessionType)
    end

    sessionType = SessionTypes[sessionType]

    local session
    -- try get the session with :GetSession
    if sessionID and sessionType.static.GetSession then
        session = sessionType.static:GetSession(GGL.sessionDb.Sessions, sessionID) --sessionDB, sessionID
    end

    -- check if the session is closes
    if session and session:IsClosed() then
        self:CloseSession(sessionType.key)
        return self:GetNewSession(sessionType.key, closeRunning)
    end

    -- check for session or create new one
    session = session or Templates:New(sessionType.name, GGL.sessionDb.Sessions)

    -- save session as running
    SessionRegister[sessionType.key] = session

    -- save the session
    GGL.sessionDb.SessionRegister[sessionType.key] = session:GetSessionID()

    return session
end

--- Gets a registered session for a sessionType if aviable
-- @param       sessionType         <string>
-- @return  sessionTemplate
function Session:GetRegisteredSession(sessionType)
    if not sessionType or not SessionRegister[sessionType] then return end

    return SessionRegister[sessionType]
end

--- Close a running sessions
-- @param       sessionType         <string>
function Session:CloseSession(sessionType)
    if not sessionType or not SessionRegister[sessionType] then return end

    SessionRegister[sessionType]:Close()
    if SessionRegister[sessionType] == StartedSession then
        self:UnlockSession()
    end
    SessionRegister[sessionType] = nil
    GGL.sessionDb.SessionRegister[sessionType] = nil
end

--- Close a running sessions
-- @param       sessionType         <string>
function Session:StopSession(sessionType)
    if not sessionType or not SessionRegister[sessionType] then return end

    SessionRegister[sessionType]:Stop()
end

--- Stops all running sessions
function Session:StopAllSessions()
    for sessionType, sessionTemplate in pairs(SessionRegister) do
        if sessionTemplate ~= StartedSession then
            self:StopSession(sessionType)
        end
    end
end

--- Close all running sessions
function Session:CloseAllSessions()
    for sessionType, sessionTemplate in pairs(SessionRegister) do
        self:CloseSession(sessionType)
    end
end

--- Start a session
-- @param       sessionData         <string> sessionType or sessionID or sessionTemplate
function Session:StartSession(sessionData)
    if not sessionData then return end

    -- get session
    local session
    if type(sessionData) == "string" then
        session = self:TryGetHostSession(sessionData)
    elseif type(sessionData) == "table" then
        session = sessionData
    else
        return false
    end

    -- save
    self:LockSession(session)

    -- start
    session:Start()
end

--- Locks a session so no other can started
-- @param       sessionData         <string> sessionType or sessionID or sessionTemplate
function Session:LockSession(sessionData)
    -- get session
    local session = GetSessionFromSessionData(sessionData)
    if not session or StartedSession == session then return end

    if StartedSession and StartedSession ~= session then
        self:UnlockSession()
    end

    -- save
    StartedSession = session

    -- lock event
    Session:FireCallbackEvent("Session_OnSessionLocked", Session, StartedSession)
end

--- Unlocks a session so no other can started
function Session:UnlockSession(session)
    if not StartedSession then return end

    if session and session ~= StartedSession then return end

    --if StartedSession:IsRunning() then
        StartedSession:Stop()
    --end

    -- save
    session = StartedSession
    StartedSession = nil

    -- release event
    Session:FireCallbackEvent("Session_OnSessionLockReleased", Session, session)
end

--- Gets the lockes session
function Session:GetLockedSession()
    return StartedSession
end

--- Checks if the player can start a new session or is host of a running
function Session:PlayerCanAccessRunningSession(sessionType)
    if StartedSession then
        if (not StartedSession.PlayerIsValidHost or StartedSession:PlayerIsValidHost()) and StartedSession:GetSessionType() == sessionType then
            return true
        else
            return false
        end
    else
        return true
    end
end

-- ###########################
-- options
-- ###########################
local OptionsBaseTable

-- Callbacks
local function CB_it_PlayerInviteStateUpdate(inviteTemplate, playerTemplate, inviteState)
    OptionsBaseTable:UpdatePlayer(playerTemplate)
end

local function CB_it_NewPlayerAdded(inviteTemplate, playerTemplate)
    OptionsBaseTable:AddPlayer(playerTemplate)
end

local function CB_it_PlayerRemoved(inviteTemplate, playerTemplate, inviteState)
    OptionsBaseTable:RemovePlayer(playerTemplate)
end

local function CB_s_OnClose(sessionTemplate)
    OptionsBaseTable:EndSession()
end

local function CB_s_OnStop(sessionTemplate)
    OptionsBaseTable:UpdateGUI(true, false)
end

local function RegisterInviteTemplateCB(inviteTemplate, register)
    if not inviteTemplate then return end
    local session = inviteTemplate:GetSession()
    if register and session then
        session:RegisterCallbackEvent("SessionTemplate_SessionClosed", CB_s_OnClose)
        session:RegisterCallbackEvent("SessionTemplate_SessionStopped", CB_s_OnStop)
    elseif session then
        session:UnregisterCallbackEvent("SessionTemplate_SessionClosed", CB_s_OnClose)
        session:UnregisterCallbackEvent("SessionTemplate_SessionStopped", CB_s_OnStop)
    end
end

-- on demand gui functions
local function Description_OnEnterPressed(self, event, text)
    OptionsBaseTable:SetDescription(text or "")
    self:ClearFocus()
end

local function AddPlayerName_OnEnterPressed(self, event, text)
    text = strtrim(text)
    if text ~= "" then
        local playerTemplate = Templates:Get("WoWPlayer", text)
        if playerTemplate then
            OptionsBaseTable:AddPlayer(playerTemplate)
        end
    end
    self:SetText()
    self:ClearFocus()
end

local function OpenBNInvites_OnClick(self, event, mb)
    OptionsBaseTable:OpenBNInvites()
end

local function OpenInvitesFrame_OnClick(self, event, mb)
    OptionsBaseTable:ShowInviteFrame()
end

local function EnableCrossClient_OnValueChanged(self, event, value)
    OptionsBaseTable.db.crossClient = value
    OptionsBaseTable:UpdateCrossClientState()
end

local function PlayerDeleteButton_OnCLick(self, event, button)
    local children = OptionsBaseTable.gui.inviteList.children
    local numChildern = #children
    if numChildern > 0 then
        local fix = 0
        for i = 1, numChildern do
            local child = children[i]
            children[i] = nil

            if child == self then
                -- get playerTemplate
                local playerTemplate = child:GetPlayer()
                -- fire player remove
                OptionsBaseTable:RemovePlayer(playerTemplate)
                -- set position fix
                fix = fix + 1
            else
                child:SetCount(i-fix)
                children[i-fix] = child
            end
        end
        OptionsBaseTable.gui.inviteList:DoLayout()
    end
end

local function UpdateInviteTemplateData()
    local session = OptionsBaseTable.session
    local inviteTemplate = OptionsBaseTable.inviteTemplate
    if not session or not inviteTemplate then return end
    local db = OptionsBaseTable.db

    -- set description
    inviteTemplate:SetDescription(db.description)
end

local OPTIONS_BASE_FUNCTIONS
OPTIONS_BASE_FUNCTIONS = {
    --- gets called when the gui release
    ["OnRelease"] = function(self)
        -- wipe options gui table
        if self.gui.guiIsSet then
            wipe(self.gui)
        end
        -- remove inviteTemplate
        RegisterInviteTemplateCB(self.inviteTemplate, false)
        self.inviteTemplate = nil
        -- unregister callbacks
        self:UnregisterAllCallbackEvents()
        -- wipe player list
        wipe(self.playerList)
        self.playerListCount = 0
        -- wipe override table
        wipe(self.__ovFunctions)

        self.db = nil
    end,

    --- overrides a function
    ["OverrideFunction"] = function(self, functionName, func)
        self.__ovFunctions[functionName] = func
    end,

    --- ignore the override
    ["RawCall"] = function(self, functionName, ...)
        OPTIONS_BASE_FUNCTIONS[functionName](self, ...)
    end,

    --- Set the db table
    -- @param   db      <table>
    ["SetDB"] = function(self, db)
        if self.db then return end
        self.db = db

        -- add players from list
        if db.inviteList then
            for playerSyncID in pairs(db.inviteList) do
                self:AddPlayer(GetPlayerFromSync(playerSyncID))
            end
        end

        self:Refresh()
    end,

    --- Set the sessionType table
    -- @param   db      <string> table:SessionTypes
    ["SetSessionType"] = function(self, sessionType)
        if sessionType and not SessionTypes[sessionType] then return end
        self.sessionType = sessionType
        if sessionType then
            local session = Session:TryGetHostSession(sessionType)
            if not session then
                error(format("Session for SessionType '%s' not found.", sessionType))
            end
            self.session = session
            if session.GetInviteTemplate then
                self:SetInviteTemplate(session:GetInviteTemplate())
            end
        else
            RegisterInviteTemplateCB(self.inviteTemplate, false)
            -- wipe player list
            wipe(self.playerList)
            self.playerListCount = 0
            -- set session data
            self.session = nil
            self.inviteTemplate = nil
        end
    end,

    --- Sets the inviteTemplate
    -- @param   inviteTemplate      <inviteTemplate>
    ["SetInviteTemplate"] = function(self, inviteTemplate)
        if self.inviteTemplate and (inviteTemplate ~= true) then return end
        self.inviteTemplate = inviteTemplate

        if not inviteTemplate then return end

        -- set callbacks
        RegisterInviteTemplateCB(self.inviteTemplate, inviteTemplate:GetBattleNetInvitesState())
    end,

    --- Refresh the options
    ["Refresh"] = function(self)
        local gui = self.gui
        local db = self.db

        -- set time
        gui.time:SetText(db.duration)

        -- set show main frame check box
        gui.isEndlessTimer:SetValue(db.isEndlessTimer)

        -- set "sets"
        gui.setDropdown:SetValue(db.mainSet or "Base")

        -- description
        if gui.descEditBox then
            gui.descEditBox:SetText(db.description)
        end

        gui.enableSpecialEvent:SetValue(db.enableSpecialEvent)
        gui.specialEventDropdown:SetValue(db.specialEvent or "Opening")
    end,

    --- removes a frame from the options
    -- @param   guiTableName        <string> name in the self.gui table
    ["RemoveFrame"] = function(self, guiTableName)
        if self.gui[guiTableName] and guiTableName ~= "topGroup" then
            self.gui[guiTableName]:Release()
            self.gui[guiTableName] = nil
        end
    end,

    --- Sets a new description and refresh description for inviteTemplate if aviable
    -- @param   description     <string>
    -- @param   forceSet        <bool> don't compare with existing description
    ["SetDescription"] = function(self, description, forceSet)
        if not forceSet and self.db.description == description then return end

        self.db.description = description

        if self.inviteTemplate then
            self.inviteTemplate:SetDescription(description)
        end
    end,

    --- Reloads the set settings
    ["ReloadSettings"] = function(self)
        if not self.session or not self.session:IsRunningTemplate() then return end

        self.session:SetData(self.db.mainSet or "Base", true)
        self.session:ReloadSet()
    end,

    --- Opens the invites for player on the BN List
    -- @param   state       <bool> open/close invites
    ["OpenBNInvites"] = function(self, state)
        if not self.inviteTemplate then return end
        self:UpdateCrossClientState()
        local itState = self.inviteTemplate:GetBattleNetInvitesState()
        -- only do something if state is different
        if itState and itState == state then return end
        -- get new state
        state = (state~=nil) and state or (not itState)
        -- lock session
        Session:LockSession(self.session)
        -- update
        UpdateInviteTemplateData()
        -- callbacks
        RegisterInviteTemplateCB(self.inviteTemplate, state)
        -- set new state
        self.inviteTemplate:OpenBattleNetInvites(state)
        -- update
        self:UpdateGUI(true, false)
        -- open invite frame
        if state then
            self:ShowInviteFrame()
        end
    end,

    ["ShowInviteFrame"] = function(self)
        if not self.inviteTemplate then return end
        GGL.GUI.Challenge:ShowInviteFrame(self.session)
    end,

    --- Start/Stop the session
    ["StartSession"] = function(self)
        if not self.sessionType then return end
        UpdateInviteTemplateData()
        -- stop session when its running
        if self.session:IsRunningTemplate() then
            Session:StopSession(self.sessionType)
        else
            self:UpdateCrossClientState()
            -- set timer
            if self.db.isEndlessTimer then
                self.session:SetTimerEndless()
            else
                self.session:SetTimerDuration(self.db.duration*60)
            end
            -- check is data is set
            if not self.session:DataIsSet() then
                self.session:SetData(self.db.mainSet or "Base")
            end
            -- set special event
            if self.db.enableSpecialEvent then
                self.session:SetSpecialEvent(self.db.specialEvent)
            end

            -- update main frame display
            Session:StartSession(self.sessionType)
        end
        self:UpdateGUI(true, false)
    end,

    --- Ends the session
    ["EndSession"] = function(self)
        if not self.sessionType then return end
        local sessionType = self.sessionType
        -- close session
        self:SetSessionType(nil)

        Session:CloseSession(sessionType)

        self:SetSessionType(sessionType)

        self:UpdateGUI(true, false)
    end,

    --- Update
    ["UpdateGUI"] = function(self, gui, inviteList)
        -- update gui
        if gui and self.gui and self.gui.statusText then
            gui = self.gui

            local session = self.session

            local statusText = gui.statusText
            local reloadSettings = gui.reloadSettings
            local startSession = gui.startSession
            local endSession = gui.endSession

            -- default states
            startSession:SetDisabled(false)
            endSession:SetDisabled(false)
            reloadSettings:SetDisabled(true)

            -- default text
            startSession:SetText(L["Start session"])
            GGL.GUI:AddToolTipForAceGUIWidget(startSession, L["Start session"], nil)


            if gui.openBNInvites and session.inviteTemplate and session.inviteTemplate:GetBattleNetInvitesState() then
                gui.openBNInvites:SetText(L["Close Battle.net invites"])
            elseif gui.openBNInvites then
                gui.openBNInvites:SetText(L["Open Battle.net invites"])
            end

            if not session then
                -- no session found
                statusText:SetText(L["No session started."])
                endSession:SetDisabled(true)
            elseif session:IsClosed() then
                -- session is closed
                statusText:SetText(L["Session is closed. Start a new one."])
                endSession:SetDisabled(true)
                self:EndSession()
            elseif session:IsRunningTemplate() then
                -- session is running
                statusText:SetText(L["Session is running."])
                startSession:SetText(L["Pause Session"])
                GGL.GUI:AddToolTipForAceGUIWidget(startSession, L["Pause Session"], L["Pause the session."], L["Paused sessions can be resumed."])
                reloadSettings:SetDisabled(false)
            elseif session:IsRunning() or session:IsStopped() then
                -- session is not running but can be resumed
                statusText:SetText(L["Session found. Send new invites or resume."])
                startSession:SetText(L["Resume session"])
                GGL.GUI:AddToolTipForAceGUIWidget(startSession, L["Resume session"], L["Resume a stopped session."])
            elseif session.invitesOpen or (session.inviteTemplate and session.inviteTemplate:GetBattleNetInvitesState()) then
                -- invites are send out
                statusText:SetText(L["Invites send."])
            elseif session then
                -- clear session found
                statusText:SetText(L["No session started."])
                --endSession:SetDisabled(true)
            else
                statusText:SetText("That should not happen?")
            end
        end

        if inviteList then
            self:SetInviteTemplate(self.inviteTemplate)
        end
    end,

    ["UpdateCrossClientState"] = function(self)
        if self.session.EnableCrossClientInvite then
            self.session:EnableCrossClientInvite(OptionsBaseTable.db.crossClient)
        end
    end,


    --- adds description editBox into the options
    ["GUI_AddDescription"] = function(self)
        if self.gui.descEditBox then return end
        local descEditBox = AceGUI:Create("MultiLineEditBox")
        descEditBox:SetFullWidth(true)
        descEditBox:SetLabel(L["Description"])
        descEditBox:SetMaxLetters(3000)
        descEditBox:SetCallback("OnEnterPressed", Description_OnEnterPressed)
        self.gui.topGroup:AddChild(descEditBox)
        self.gui.descEditBox = descEditBox

        self:Refresh()
    end,

    --- adds battle.net invite button
    ["GUI_AddBattleNetInvite"] = function(self)
        if self.gui.openBNInvites then return end
        local openBNInvites = AceGUI:Create("Button")
        openBNInvites:SetText(L["Open Battle.net invites"])
        GGL.GUI:AddToolTipForAceGUIWidget(openBNInvites, L["Open Battle.net invites"], L["Opens invite for all battle.net friends."], L["This invites all battle.net friends as long it stays open."])
        openBNInvites:SetAutoWidth(true)
        openBNInvites:SetCallback("OnClick", OpenBNInvites_OnClick)
        self.gui.bottomExtraGroup:AddChild(openBNInvites)
        self.gui.openBNInvites = openBNInvites

        local openInvitesFrame = AceGUI:Create("Button")
        openInvitesFrame:SetText(L["Open Invite list"])
        GGL.GUI:AddToolTipForAceGUIWidget(openInvitesFrame, L["Open Invite list"])
        openInvitesFrame:SetAutoWidth(true)
        openInvitesFrame:SetCallback("OnClick", OpenInvitesFrame_OnClick)
        self.gui.bottomExtraGroup:AddChild(openInvitesFrame)
        self.gui.openInvitesFrame = openInvitesFrame

        local enableCrossClient = AceGUI:Create("CheckBox")
        enableCrossClient:SetLabel(L["Cross client"])
        GGL.GUI:AddToolTipForAceGUIWidget(enableCrossClient, L["Cross client"], L["Allows inviting friends from other game versions."])
        enableCrossClient:SetValue(OptionsBaseTable.db.crossClient)
        enableCrossClient:SetCallback("OnValueChanged", EnableCrossClient_OnValueChanged)
        self.gui.bottomExtraGroup:AddChild(enableCrossClient)
        self.gui.enableCrossClient = enableCrossClient

        self:UpdateCrossClientState()

        self:UpdateGUI(true, false)
    end,

    --- adds "Add Player" edit box
    ["GUI_AddInviteByName"] = function(self)
        local addPlayerName = AceGUI:Create("EditBox")
        addPlayerName:SetWidth(200)
        addPlayerName:SetLabel(L["Add Player"])
        addPlayerName:SetCallback("OnEnterPressed", AddPlayerName_OnEnterPressed)
        self.gui.bottomExtraGroup:AddChild(addPlayerName)
        self.gui.addPlayerName = addPlayerName
    end,
}

-- gets the options base metatable
local function Options_GetBaseTable()
    if not OptionsBaseTable then
        local base = {
            gui = {},
            playerList = {},
            playerListCount = 0,
            inviteTemplate = nil,
            __ovFunctions = {},
        }

        -- callback handler
        GGL:AddCallbackHandler(base)

        setmetatable(base, {__index = function(t,k)
            return t.__ovFunctions[k] or OPTIONS_BASE_FUNCTIONS[k]
        end,

        OPTIONS_BASE_FUNCTIONS})
        OptionsBaseTable = base
    end
    return OptionsBaseTable
end

local function WidgetOnRelease(self, event)
    OptionsBaseTable:OnRelease()
end

local function GetSetDropDownList(setType)
    local db = GGL.db.Config.Sets[setType]

    local list = {}
    for k,v in pairs(db) do
        list[k] = v.name
    end

    return list
end

-- ###AceGUI
local function TimeEdit_OnEnterPressed(self, event, text)
    text = tonumber(text)
    if not text then
        self:SetText(OptionsBaseTable.db.duration or 60)
    else
        if text > 6060 then
            text = 6060
            self:SetText(text)
        end
        OptionsBaseTable.db.duration = text
    end
    self:ClearFocus()
end

local function SetDropDown_OnValueChanged(self, event, value)
    OptionsBaseTable.db.mainSet = value or "Base"
end

local function ReloadSettings_OnClick(self, event, mb)
    OptionsBaseTable:ReloadSettings()
end

local function IsEndlessTimer_OnValueChanged(self, event, state)
    OptionsBaseTable.db.isEndlessTimer = state
end

local function EnableSpecialEvent_OnValueChanged(self, event, state)
    OptionsBaseTable.db.enableSpecialEvent = state
end

local function SpecialEventDropDown_OnValueChanged(self, event, value)
    OptionsBaseTable.db.specialEvent = value
end

local function StartSession_OnClick(self, event, mb)
    OptionsBaseTable:StartSession()
end

local function EndSession_OnClick(self, event, mb)
    OptionsBaseTable:EndSession()
end

--- Creates the gui
local function Options_CreateGUI(targetGroup, guiTable)
    if guiTable.guiIsSet then return end

    -- scroll
    local scroll = AceGUI:Create("ScrollFrame")
    scroll:SetLayout("Flow")
    targetGroup:AddChild(scroll)
    guiTable.scroll = scroll

    -- ## top Group
    local topGroup = AceGUI:Create("SimpleGroup")
    topGroup:SetFullWidth(true)
    topGroup:SetLayout("Flow")
    topGroup:SetCallback("OnRelease", WidgetOnRelease)
    scroll:AddChild(topGroup)
    guiTable.topGroup = topGroup

    local time = AceGUI:Create("EditBox")
    time:SetWidth(80)
    time:SetLabel(L["Time"])
    time:SetMaxLetters(4)
    time:SetCallback("OnEnterPressed", TimeEdit_OnEnterPressed)
    topGroup:AddChild(time)
    guiTable.time = time

    local timeText = AceGUI:Create("Label")
    timeText:SetText(L["min"])
    topGroup:AddChild(timeText)
    guiTable.timeText = timeText

    local isEndlessTimer = AceGUI:Create("CheckBox")
    isEndlessTimer:SetLabel(L["Endless session"])
    isEndlessTimer:SetCallback("OnValueChanged", IsEndlessTimer_OnValueChanged)
    topGroup:AddChild(isEndlessTimer)
    guiTable.isEndlessTimer = isEndlessTimer

    -- # sets
    local settingsLine = AceGUI:Create("Heading")
    settingsLine:SetText(L["Settings"])
    settingsLine:SetFullWidth(true)
    topGroup:AddChild(settingsLine)
    guiTable.settingsLine = settingsLine

    local setDropdown = AceGUI:Create("Dropdown")
    setDropdown:SetList(GetSetDropDownList(GGL.SET_ID_LIST.Set))
    setDropdown:SetLabel(L["Filter Set"])
    setDropdown:SetCallback("OnValueChanged", SetDropDown_OnValueChanged)
    topGroup:AddChild(setDropdown)
    guiTable.setDropdown = setDropdown

    local reloadSettings = AceGUI:Create("Button")
    reloadSettings:SetText(L["Reload settings"])
    reloadSettings:SetAutoWidth(true)
    reloadSettings:SetCallback("OnClick", ReloadSettings_OnClick)
    GGL.GUI:AddToolTipForAceGUIWidget(reloadSettings, L["Reload settings"], L["Reloads the settings for all players."], L["This also refresh all lists."])
    topGroup:AddChild(reloadSettings)
    guiTable.reloadSettings = reloadSettings

    local sessionLine = AceGUI:Create("Heading")
    --sessionLine:SetText(L["Session"])
    sessionLine:SetFullWidth(true)
    scroll:AddChild(sessionLine)
    guiTable.sessionLine = sessionLine

    -- ## bottom extra group
    local bottomExtraGroup = AceGUI:Create("SimpleGroup")
    bottomExtraGroup:SetFullWidth(true)
    bottomExtraGroup:SetLayout("Flow")
    scroll:AddChild(bottomExtraGroup)
    guiTable.bottomExtraGroup = bottomExtraGroup

    -- ## bottom group
    local bottomGroup = AceGUI:Create("SimpleGroup")
    bottomGroup:SetFullWidth(true)
    bottomGroup:SetLayout("Flow")
    scroll:AddChild(bottomGroup)
    guiTable.bottomGroup = bottomGroup

    local startSession = AceGUI:Create("Button")
    startSession:SetText(L["Start session"])
    startSession:SetAutoWidth(true)
    startSession:SetCallback("OnClick", StartSession_OnClick)
    bottomGroup:AddChild(startSession)
    guiTable.startSession = startSession

    local endSession = AceGUI:Create("Button")
    endSession:SetText(L["End running session"])
    GGL.GUI:AddToolTipForAceGUIWidget(endSession, L["End running session"], L["Ends the running session."])
    endSession:SetAutoWidth(true)
    endSession:SetCallback("OnClick", EndSession_OnClick)
    bottomGroup:AddChild(endSession)
    guiTable.endSession = endSession

    -- ## status group
    local statusGroup = AceGUI:Create("SimpleGroup")
    statusGroup:SetFullWidth(true)
    statusGroup:SetLayout("Flow")
    scroll:AddChild(statusGroup)
    guiTable.statusGroup = statusGroup

    local statusText = AceGUI:Create("Label")
    statusText:SetFullWidth(true)
    statusText:SetHeight(15)
    statusText:SetText("Status")
    statusGroup:AddChild(statusText)
    guiTable.statusText = statusText

    local specialEventLine = AceGUI:Create("Heading")
    specialEventLine:SetText(L["Special event"])
    specialEventLine:SetFullWidth(true)
    scroll:AddChild(specialEventLine)
    guiTable.specialEventLine = specialEventLine

    -- ## special group
    local specialGroup = AceGUI:Create("SimpleGroup")
    specialGroup:SetFullWidth(true)
    specialGroup:SetLayout("Flow")
    scroll:AddChild(specialGroup)
    guiTable.specialGroup = specialGroup

    local enableSpecialEvent = AceGUI:Create("CheckBox")
    enableSpecialEvent:SetWidth(30)
    enableSpecialEvent:SetCallback("OnValueChanged", EnableSpecialEvent_OnValueChanged)
    specialGroup:AddChild(enableSpecialEvent)
    guiTable.enableSpecialEvent = enableSpecialEvent

    local specialEventDropdown = AceGUI:Create("Dropdown")
    specialEventDropdown:SetList(SPECIAL_EVENT_DD_LIST)
    --specialEventDropdown:SetLabel(L["Filter Set"])
    specialEventDropdown:SetCallback("OnValueChanged", SpecialEventDropDown_OnValueChanged)
    specialGroup:AddChild(specialEventDropdown)
    guiTable.specialEventDropdown = specialEventDropdown

    guiTable.guiIsSet = true
end

local function Options_CreateLockedGUI(targetGroup, guiTable)
    if guiTable.guiIsSet then return end

    local lockedGroup = AceGUI:Create("SimpleGroup")
    lockedGroup:SetFullWidth(true)
    lockedGroup:SetFullHeight(true)
    lockedGroup:SetLayout("Fill")
    targetGroup:AddChild(lockedGroup)
    lockedGroup:SetCallback("OnRelease", WidgetOnRelease)
    guiTable.lockedGroup = lockedGroup

    local lockedText = AceGUI:Create("Label")
    lockedText:SetFullWidth(true)
    lockedText:SetFullHeight(true)
    lockedText:SetText(L["You can't start a new session as long one is running."])
    lockedGroup:AddChild(lockedText)
    guiTable.lockedText = lockedText

    guiTable.guiIsSet = true
end

--- Add the gui into the options
-- @param   targetGroup     <AceGUI-group>
function Session:Options_Add(targetGroup, sessionType, dbTable)
    -- get options table
    local options = Options_GetBaseTable()

    -- fill gui
    if Session:PlayerCanAccessRunningSession(sessionType) then
        Options_CreateGUI(targetGroup, options.gui)
        options:SetSessionType(sessionType)
        options:SetDB(dbTable)

        return options
    else
        Options_CreateLockedGUI(targetGroup, options.gui)

        return false
    end
end

--- Try update the session UI of open
function Session:Options_Update(sessionType)
    if sessionType and OptionsBaseTable and OptionsBaseTable.sessionType == sessionType then
        UpdateInviteTemplateData()
        OptionsBaseTable:UpdateGUI(true, false)
    end
end